import numpy as np
import jax.numpy as jnp
import jax
import tqdm


def log_bayes_factor_score(
    log_marginal_likelihood_1: float,
    log_marginal_likelihood_2: float,
) -> float:
    """
    Calculate the log bayes factor from two log
    marginal likelihoods.

    Arguments
    ----------
    log_marginal_likelihood_1 : float
        The log marginal likelihood of the first model.

    log_marginal_likelihood_2 : float
        The log marginal likelihood of the second model.

    Returns
    --------

    bayes_factor : float
        The bayes factor between the two models.

    """
    return log_marginal_likelihood_1 - log_marginal_likelihood_2


def bayes_factor_score(
    log_marginal_likelihood_1: float,
    log_marginal_likelihood_2: float,
) -> float:
    """
    Calculate the bayes factor from two log
    marginal likelihoods.

    If the log of the bayes factor is greater than 1e2, it will
    return infinity. This is to prevent numerical issues.

    Arguments
    ----------
    log_marginal_likelihood_1 : float
        The log marginal likelihood of the first model.

    log_marginal_likelihood_2 : float
        The log marginal likelihood of the second model.

    Returns
    --------

    bayes_factor : float
        The bayes factor between the two models.

    """

    log_bayes_factor_value = log_bayes_factor_score(
        log_marginal_likelihood_1,
        log_marginal_likelihood_2,
    )

    if log_bayes_factor_value > 1e2:
        return np.inf

    return np.exp(log_bayes_factor_value)


def negative_log_likelihood_jax(y: jnp.ndarray, p: jnp.ndarray) -> jnp.ndarray:
    """

    The negative log-likelihood of the model.

    Parameters
    ----------

    y: jnp.ndarray
        The true labels of the data. The
        shape of the array should be the same as
        the shape of the array :code:`p`.

    p: jnp.ndarray
        The predicted probabilities of the data. The
        shape of the array should be the same as
        the shape of the array :code:`y`.

    Returns
    -------

    jnp.ndarray
        The negative log-likelihood of the model. The shape of the array
        will be the same as the shape of the array :code:`y` and :code:`p`.

    """
    return -(y * jnp.log(p) + (1 - y) * jnp.log1p(-p))


def mean_pairwise_euclidean_distance(X, Y):
    """
    The mean pairwise Euclidean distance between two sets of points.

    Parameters
    ----------

    X: np.ndarray
        The first set of points. The shape of the array should be
        (n_samples, n_features).

    Y: np.ndarray
        The second set of points. The shape of the array should be
        (n_samples, n_features).


    Returns
    -------

    float
        The mean pairwise Euclidean distance between the two sets of points.


    """

    X, Y = jnp.array(X), jnp.array(Y)

    # Inner loop to compute pairwise distances between x and all y
    inner_loop = lambda x, Y: jax.vmap(
        lambda y: jnp.linalg.norm(x - y, ord=2), in_axes=0
    )(Y)
    jitted_inner_loop = jax.jit(inner_loop)

    # Outer loop to compute pairwise distances for all x
    outer_loop = jax.vmap(lambda x: jitted_inner_loop(x, Y), in_axes=0)(X)

    # Return the mean of all pairwise distances
    return jnp.mean(outer_loop)


def mean_pairwise_euclidean_distance_chunked(
    X,
    Y,
    chunk_size=100,
    verbose=False,
):
    """
    The mean pairwise Euclidean distance between two sets of points.
    This function computes the pairwise distances in chunks to reduce
    memory usage.


    Parameters
    ----------

    X: np.ndarray
        The first set of points. The shape of the array should be
        (n_samples, n_features).

    Y: np.ndarray
        The second set of points. The shape of the array should be
        (n_samples, n_features).


    Returns
    -------

    float
        The mean pairwise Euclidean distance between the two sets of points.

    """

    def compute_chunked_distances(x_chunk, y_chunk):
        # Use vmap to compute pairwise distances for the chunk
        # similar to mean_pairwise_euclidean_distance
        return jax.vmap(
            lambda x: jax.vmap(lambda y: jnp.linalg.norm(x - y, ord=2))(y_chunk)
        )(x_chunk)

    compute_chunked_distances = jax.jit(compute_chunked_distances)

    total_sum = 0.0

    X, Y = jnp.array(X), jnp.array(Y)

    # Iterate over chunks of X
    for i in tqdm.trange(0, X.shape[0], chunk_size, disable=not verbose):
        x_chunk = X[i : i + chunk_size]

        # Iterate over chunks of Y
        for j in range(0, Y.shape[0], chunk_size):
            y_chunk = Y[j : j + chunk_size]

            # Compute pairwise distances for the current chunk
            chunk_distances = compute_chunked_distances(x_chunk, y_chunk)

            # Sum the distances for this chunk
            total_sum += jnp.sum(chunk_distances) / Y.shape[0]

    return total_sum / X.shape[0]


def energy_distance(X, Y):
    """
    The energy distance between two sets of points.

    https://en.wikipedia.org/wiki/Energy_distance


    Parameters
    ----------

    X: np.ndarray
        The first set of points. The shape of the array should be
        (n_samples, n_features).

    Y: np.ndarray
        The second set of points. The shape of the array should be
        (n_samples, n_features).


    Returns
    -------

    float
        The energy distance between the two sets of points.

    """

    X, Y = jnp.array(X), jnp.array(Y)

    return (
        2 * mean_pairwise_euclidean_distance(X, Y)
        - mean_pairwise_euclidean_distance(X, X)
        - mean_pairwise_euclidean_distance(Y, Y)
    )


def energy_distance_chunked(X, Y, chunk_size=1000):
    """
    The energy distance between two sets of points.
    This function computes the pairwise distances in chunks to reduce
    memory usage.

    https://en.wikipedia.org/wiki/Energy_distance


    Parameters
    ----------

    X: np.ndarray
        The first set of points. The shape of the array should be
        (n_samples, n_features).

    Y: np.ndarray
        The second set of points. The shape of the array should be
        (n_samples, n_features).


    Returns
    -------

    float
        The energy distance between the two sets of points.

    """

    X, Y = jnp.array(X), jnp.array(Y)

    Exy = mean_pairwise_euclidean_distance_chunked(X, Y, chunk_size=chunk_size)
    Exx = mean_pairwise_euclidean_distance_chunked(X, X, chunk_size=chunk_size)
    Eyy = mean_pairwise_euclidean_distance_chunked(Y, Y, chunk_size=chunk_size)

    return jnp.sqrt(2 * Exy - Exx - Eyy)


def e_coefficient(X, Y, chunk_size=1000):
    """
    The e coefficient between two sets of points.
    This function computes the pairwise distances in chunks to reduce
    memory usage.

    https://en.wikipedia.org/wiki/Energy_distance


    Parameters
    ----------

    X: np.ndarray
        The first set of points. The shape of the array should be
        (n_samples, n_features).

    Y: np.ndarray
        The second set of points. The shape of the array should be
        (n_samples, n_features).


    Returns
    -------

    float
        The e-coefficient between the two sets of points.

    """

    Exy = mean_pairwise_euclidean_distance_chunked(X, Y, chunk_size=chunk_size)
    Exx = mean_pairwise_euclidean_distance_chunked(X, X, chunk_size=chunk_size)
    Eyy = mean_pairwise_euclidean_distance_chunked(Y, Y, chunk_size=chunk_size)

    return (2 * Exy - Exx - Eyy) / (2 * Exy)


def levenshtein_score(s1, s2, key=hash):
    """Levenshtein string distance with edits."""

    # Based on
    # https://github.com/interpretml/LLM-Tabular-Memorization-Checker/blob/d2252675a3ff9ac1d0887ca920d7d6fe347b3f66/tabmemcheck/utils.py#L614
    # Generate the cost matrix for the two strings
    def costmatrix(s1, s2, key=hash):
        rows = []

        previous_row = range(len(s2) + 1)
        rows.append(list(previous_row))

        for i, c1 in enumerate(s1):
            current_row = [i + 1]
            for j, c2 in enumerate(s2):
                insertions = previous_row[j + 1] + 1
                deletions = current_row[j] + 1
                substitutions = previous_row[j] + (key(c1) != key(c2))
                current_row.append(min(insertions, deletions, substitutions))
            previous_row = current_row

            rows.append(previous_row)

        return rows

    rows = costmatrix(s1, s2, key)

    return rows[-1][-1]



import matplotlib.pyplot as plt
import seaborn as sns
import contextlib
import matplotlib
import numpy as np
from cycler import cycler
import typing as t


def save_fig(fig: plt.figure, file_name: str, **kwargs) -> None:
    """
    This function saves a pdf, png, and svg of the figure,
    with :code:`dpi=300`.


    Arguments
    ---------

    - fig: plt.figure:
        The figure to save.

    - file_name: str:
        The file name, including path, to save the figure at.
        This should not include the extension, which will
        be added when each file is saved.

    """

    fig.savefig(f"{file_name}.pdf", **kwargs)
    fig.savefig(f"{file_name}.png", dpi=300, **kwargs)
    fig.savefig(f"{file_name}.svg", **kwargs)


# colours
tol_muted = [
    "#332288",
    "#88CCEE",
    "#44AA99",
    "#117733",
    "#999933",
    "#DDCC77",
    "#CC6677",
    "#882255",
    "#AA4499",
]

ibm = [
    "#648fff",
    "#fe6100",
    "#dc267f",
    "#785ef0",
    "#ffb000",
]


# colour map
def set_colour_map(colours: list = tol_muted):
    """
    Sets the default colour map for all plots.



    Examples
    ---------

    The following sets the colourmap to :code:`tol_muted`:

    .. code-block::

        >>> set_colour_map(colours=tol_muted)


    Arguments
    ---------

    - colours: list, optional:
        Format that is accepted by
        :code:`cycler.cycler`.
        Defaults to :code:`tol_muted`.

    """
    custom_params = {"axes.prop_cycle": cycler(color=colours)}
    matplotlib.rcParams.update(**custom_params)


# context functions
@contextlib.contextmanager
def temp_colour_map(colours=tol_muted):
    """
    Temporarily sets the default colour map for all plots.


    Examples
    ---------

    The following sets the colourmap to :code:`tol_muted` for
    the plotting done within the context:

    .. code-block::

        >>> with set_colour_map(colours=tol_muted):
        ...     plt.plot(x,y)


    Arguments
    ---------

    - colours: list, optional:
        Format that is accepted by
        :code:`cycler.cycler`.
        Defaults to :code:`tol_muted`.

    """
    set_colour_map(colours=colours)


@contextlib.contextmanager
def graph_theme(colours: t.List[str] = ibm, **kwargs):
    """
    Temporarily sets the default theme for all plots.


    Examples
    ---------

    .. code-block::

        >>> with graph_theme():
        ...     plt.plot(x,y)


    Arguments
    ---------

    - colours: t.List[str], optional:
        Any acceptable list to :code:`cycler`.
        Defaults to :code:`ibm`.


    """
    with matplotlib.rc_context():
        sns.set(context="paper", style="whitegrid")
        custom_params = {
            "axes.spines.right": False,
            "axes.spines.top": False,
            "axes.edgecolor": "black",
            "axes.linewidth": 1,
            "axes.grid": True,
            "axes.axisbelow": True,
            "axes.prop_cycle": cycler(color=colours),
            "grid.alpha": 0.5,
            "grid.color": "#b0b0b0",
            "grid.linestyle": "--",
            "grid.linewidth": 1,
            # following requres latex
            # "font.family": "serif",
            # "font.serif": "Computer Modern",
            # "text.usetex": True,
            # end requiring latex
            "xtick.major.width": 1,
            "ytick.major.width": 1,
            "boxplot.whiskerprops.linestyle": "-",
            "boxplot.whiskerprops.linewidth": 1,
            "boxplot.whiskerprops.color": "black",
            "boxplot.boxprops.linestyle": "-",
            "boxplot.boxprops.linewidth": 1,
            "boxplot.boxprops.color": "black",
            "boxplot.meanprops.markeredgecolor": "black",
            "boxplot.capprops.color": "black",
            "boxplot.capprops.linestyle": "-",
            "boxplot.capprops.linewidth": 1.0,
        }

        matplotlib.rcParams.update(**custom_params)
        matplotlib.rcParams.update(**kwargs)

        yield



import urllib.request
import zipfile
import shutil
from pathlib import Path
import pandas as pd
import numpy as np
import sklearn.datasets as skd
from sklearn.utils import Bunch
from ucimlrepo import fetch_ucirepo
import typing
import tempfile

temp_dir = Path(tempfile.gettempdir(), "autoelicit_datasets")


uci_ids = {
    "adult": 2,
    "heart_disease": 45,
    "iris": 53,
    "wine_quality": 186,
    "wine": 109,
}


def load_and_save_uci_dataset(
    dataset_name: str,
    path: Path = temp_dir,
) -> pd.DataFrame:
    """
    This function loads and saves
    a dataset from the UCI repository
    if it has not been saved before.


    Arguments
    ---------

    dataset_name: str
        The name of the dataset to load.

    path: str or Path
        The directory where the file should be saved.
        Defaults to :code:`temp_dir`.


    Returns
    -------

    pd.DataFrame
        The dataset. The target column is named :code:`target`.


    """
    file_path = path.joinpath(f"{dataset_name}.parquet")

    if file_path.exists():
        data = pd.read_parquet(file_path)
    else:
        dataset = fetch_ucirepo(id=uci_ids.get(dataset_name, dataset_name))
        # saving the data for
        # the next time it is loaded
        print("Saving the data")
        data = dataset.data.features.assign(target=dataset.data.targets)
        data.to_parquet(file_path)

    # saving the original dataset as well used in the
    # load_and_save_uci_original_dataset function
    original_file_path = path.joinpath(f"{dataset_name}_original.parquet")
    if not original_file_path.exists():
        original_data = dataset.data.original
        original_data.to_parquet(original_file_path)

    return data


def load_and_save_uci_original_dataset(
    dataset_name: str,
    path: Path = temp_dir,
) -> pd.DataFrame:
    """
    This function loads and saves
    a dataset from the UCI repository
    if it has not been saved before.

    In particular, it saves the original dataset,
    which may include a different target column name.


    Arguments
    ---------

    dataset_name: str
        The name of the dataset to load.

    path: str or Path
        The directory where the file should be saved.
        Defaults to :code:`temp_dir`.


    Returns
    -------

    pd.DataFrame
        The dataset, with the original target column name.


    """
    file_path = path.joinpath(f"{dataset_name}_original.parquet")

    if file_path.exists():
        data = pd.read_parquet(file_path)
    else:
        dataset = fetch_ucirepo(id=uci_ids[dataset_name])
        # saving the data for
        # the next time it is loaded
        print("Saving the data")
        data = dataset.data.original
        data.to_parquet(file_path)

    # saving the standard dataset as well used in the
    # load_and_save_uci_dataset function
    standard_file_path = path.joinpath(f"{dataset_name}.parquet")
    if not standard_file_path.exists():
        standard_data = dataset.data.features.assign(target=dataset.data.targets)
        standard_data.to_parquet(standard_file_path)

    return data


def load_breast_cancer(
    return_X_y: bool = False,
    as_frame: bool = False,
) -> Bunch:
    """
    Load and return the breast cancer dataset.
    Note that the labels of this dataset are flipped compared to
    the sklearn dataset so that malignant=1 and benign=0. This
    is because it makes more sense for calculating recall and precision.

    Arguments
    ---------

    return_X_y: bool
        Whether to return data and target as numpy arrays.
        Defaults to :code:`False`.

    as_frame: bool
        Whether to return the data as a pandas DataFrame.
        Defaults to :code:`False`.


    Returns
    -------

    Bunch
        Dictionary-like object, the interesting attributes are:
        'data', the data to learn, 'target', the classification target for each sample,
        'feature_names', the feature names, 'target_names', the target names,
        and 'frame', the pandas DataFrame.

    Tuple
        If :code:`return_X_y=True`, then a tuple of data and target.
        If :code:`as_frame=True`, then as pandas DataFrame otherwise as numpy arrays.


    """

    breast_cancer_dataset = skd.load_breast_cancer(as_frame=True)

    data, target, target_names, feature_names = (
        breast_cancer_dataset["data"],
        # flipping labels so malignant=1 and benign=0 as this makes more sense for
        # calculating recall and precision
        1 - breast_cancer_dataset["target"].astype(int),
        ["benign", "malignant"],
        breast_cancer_dataset.feature_names,
    )

    if not as_frame:
        data = data.to_numpy()
        target = target.to_numpy()

    if return_X_y:
        return data, target

    breast_cancer_dataset = Bunch(
        data=data,
        target=target,
        feature_names=np.array(feature_names),
        frame=breast_cancer_dataset["data"].assign(target=target),
        target_names=np.array(target_names),
    )

    return breast_cancer_dataset


def load_california_housing(
    path: Path = temp_dir,
    return_X_y: bool = False,
    as_frame: bool = False,
) -> Bunch:
    """
    Load and return the california_housing dataset.

    Arguments
    ---------

    path: str or Path
        Path to the file where the data is stored or should be
        downloaded.
        Defaults to :code:`temp_dir`.

    return_X_y: bool
        Whether to return data and target as numpy arrays.
        Defaults to :code:`False`.

    as_frame: bool
        Whether to return the data as a pandas DataFrame.
        Defaults to :code:`False`.


    Returns
    -------

    Bunch
        Dictionary-like object, the interesting attributes are:
        'data', the data to learn, 'target', the classification target for each sample,
        'feature_names', the feature names, 'target_names', the target names,
        and 'frame', the pandas DataFrame.

    Tuple
        If :code:`return_X_y=True`, then a tuple of data and target.
        If :code:`as_frame=True`, then as pandas DataFrame otherwise as numpy arrays.


    """

    if not path.joinpath("cal_housing_py3.pkz").exists():
        print("Saving the data")

    california_housing_dataset = skd.fetch_california_housing(
        data_home=path, as_frame=True
    )

    feature_name_map = {
        "MedInc": "median income in block",
        "HouseAge": "median house age in block",
        "AveRooms": "average number of rooms in block",
        "AveBedrms": "average number of bedrooms in block",
        "Population": "block population",
        "AveOccup": "average house occupancy in block",
        "Latitude": "house block latitude",
        "Longitude": "house block longitude",
    }

    data, target, feature_names, target_names, frame = (
        california_housing_dataset["data"],
        california_housing_dataset["target"],
        california_housing_dataset["feature_names"],
        "medium house value (in $100k)",
        (
            california_housing_dataset["frame"]
            .rename(columns=feature_name_map)
            .rename(columns={"MedHouseVal": "medium house value (in $100k)"})
        ),
    )

    feature_names = [feature_name_map[fn] for fn in feature_names]

    if not as_frame:
        data = data.to_numpy()
        target = target.to_numpy()

    if return_X_y:
        return data, target

    california_housing_dataset = Bunch(
        data=data,
        target=target,
        feature_names=np.array(feature_names),
        target_names=np.array([target_names]),
        frame=frame,
    )

    return california_housing_dataset


def load_heart_disease(
    path: Path = temp_dir,
    return_X_y: bool = False,
    as_frame: bool = False,
):
    """
    Load and return the heart disease dataset: DOI: 10.24432/C52P4X

    Arguments
    ---------

    path: str or Path
        Path to the file where the data is stored or should be
        downloaded.
        Defaults to :code:`temp_dir`.

    return_X_y: bool
        Whether to return data and target as numpy arrays.
        Defaults to :code:`False`.

    as_frame: bool
        Whether to return the data as a pandas DataFrame.
        Defaults to :code:`False`.


    Returns
    -------

    Bunch
        Dictionary-like object, the interesting attributes are:
        'data', the data to learn, 'target', the classification target for each sample,
        'feature_names', the feature names, 'target_names', the target names,
        and 'frame', the pandas DataFrame.

    Tuple
        If :code:`return_X_y=True`, then a tuple of data and target.
        If :code:`as_frame=True`, then as pandas DataFrame otherwise as numpy arrays.


    """

    features_to_include = [
        "age",
        "sex",
        "trestbps",
        "chol",
        "fbs",
        "restecg",
        "thalach",
        "exang",
        "oldpeak",
        "ca",
    ]

    data = load_and_save_uci_dataset(dataset_name="heart_disease", path=path)
    y = (data["target"] > 0).astype(int).squeeze()
    data = data.drop(columns=["target"])

    # selecting only the features we want
    data = data[features_to_include]

    # processing restecg - 0 = normal, 1 = abnormal
    data = data.assign(restecg=(data["restecg"] > 0).astype(int))

    # renaming features with descriptive names

    features_to_rename = {
        "age": "age",
        "sex": "sex (1 = male, 0 = female)",
        "trestbps": "resting blood pressure (on admission to the hospital)",
        "chol": "serum cholestoral in mg/dl",
        "fbs": "fasting blood sugar > 120 mg/dl (1 = true, 0 = false)",
        "restecg": "resting electrocardiographic results (1 = abnormal, 0 = normal)",
        "thalach": "maximum heart rate achieved",
        "exang": "exercise induced angina",
        "oldpeak": "ST depression induced by exercise relative to rest",
        "ca": "number of major vessels (0 - 3) colored by flourosopy",
    }

    ## renaming the columns
    data = data.rename(columns=features_to_rename)

    # remove rows with missing values in targets or features
    missing = y.isnull().values | data.isnull().any(axis=1).values
    data = data[~missing]
    y = y[~missing]

    # collecting information to return
    data, target, frame, target_names, feature_names, categorical_features = (
        data,
        y,
        data.assign(target=y),
        np.array(["no heart disease", "heart disease"]),
        data.columns.to_numpy(),
        [1, 4, 5, 7, 9],
    )

    if not as_frame:
        data = data.to_numpy()
        target = y.to_numpy()

    if return_X_y:
        return data, target

    heart_disease_dataset = Bunch(
        data=data,
        target=target,
        feature_names=feature_names,
        frame=frame,
        target_names=target_names,
        categorical_features=categorical_features,
    )

    return heart_disease_dataset


def load_wine_quality(
    path: Path = temp_dir,
    return_X_y: bool = False,
    as_frame: bool = False,
):
    """
    Load and return the wine quality dataset: DOI: 10.24432/C56S3T

    Arguments
    ---------

    path: str or Path
        Path to the file where the data is stored or should be
        downloaded.
        Defaults to :code:`temp_dir`.

    return_X_y: bool
        Whether to return data and target as numpy arrays.
        Defaults to :code:`False`.

    as_frame: bool
        Whether to return the data as a pandas DataFrame.
        Defaults to :code:`False`.


    Returns
    -------

    Bunch
        Dictionary-like object, the interesting attributes are:
        'data', the data to learn, 'target', the classification target for each sample,
        'feature_names', the feature names, 'target_names', the target names,
        and 'frame', the pandas DataFrame.

    Tuple
        If :code:`return_X_y=True`, then a tuple of data and target.
        If :code:`as_frame=True`, then as pandas DataFrame otherwise as numpy arrays.


    """

    data = load_and_save_uci_dataset(dataset_name="wine_quality", path=path)
    # making sure targets are binary
    y = (data["target"] > 5).astype(int).squeeze()
    data = data.drop(columns=["target"])

    features_to_rename = {
        "fixed_acidity": "fixed acidity",
        "volatile_acidity": "volatile acidity",
        "citric_acid": "citric acid",
        "residual_sugar": "residual sugar",
        "chlorides": "chlorides",
        "free_sulfur_dioxide": "free sulfur dioxide",
        "total_sulfur_dioxide": "total sulfur dioxide",
        "density": "density",
        "pH": "pH",
        "sulphates": "sulphates",
        "alcohol": "alcohol",
    }

    data = data.rename(columns=features_to_rename)

    # remove rows with missing values in targets or features
    missing = y.isnull().values | data.isnull().any(axis=1).values
    data = data[~missing]
    y = y[~missing]

    # collecting information to return
    data, target, frame, target_names, feature_names = (
        data,
        y,
        data.assign(target=y),
        np.array(["bad quality", "good quality"]),
        data.columns.to_numpy(),
    )

    if not as_frame:
        data = data.to_numpy()
        target = y.to_numpy()

    if return_X_y:
        return data, target

    wine_quality_dataset = Bunch(
        data=data,
        target=target,
        feature_names=feature_names,
        frame=frame,
        target_names=target_names,
    )

    return wine_quality_dataset


def load_sk_diabetes(
    return_X_y: bool = False,
    as_frame: bool = False,
) -> Bunch:
    """
    Load and return the diabetes dataset
    (https://www4.stat.ncsu.edu/~boos/var.select/diabetes.html).

    Arguments
    ---------

    return_X_y: bool
        Whether to return data and target as numpy arrays.
        Defaults to :code:`False`.

    as_frame: bool
        Whether to return the data as a pandas DataFrame.
        Defaults to :code:`False`.


    Returns
    -------

    Bunch
        Dictionary-like object, the interesting attributes are:
        'data', the data to learn, 'target', the classification target for each sample,
        'feature_names', the feature names, 'target_names', the target names,
        and 'frame', the pandas DataFrame.

    Tuple
        If :code:`return_X_y=True`, then a tuple of data and target.
        If :code:`as_frame=True`, then as pandas DataFrame otherwise as numpy arrays.


    """

    features_to_rename = {
        "age": "age",
        "sex": "sex",
        "bmi": "body mass index",
        "bp": "average blood pressure",
        "s1": "total serum cholesterol",
        "s2": "low-density lipoproteins",
        "s3": "high-density lipoproteins",
        "s4": "total cholesterol / HDL",
        "s5": "log of serum triglycerides level",
        "s6": "blood sugar level",
    }

    diabetes_dataset = skd.load_diabetes(as_frame=True, scaled=False)

    frame = diabetes_dataset.frame.rename(columns=features_to_rename)
    data = frame.drop(columns=["target"])
    target = (
        10 * (frame["target"] - 25) / (346 - 25)
    )  # range of possible y values -> 0 to 10

    data, target, target_names, feature_names = (
        data,
        target,
        [
            "quantitative measure of diabetes disease progression one year after baseline (0 to 10)"
        ],
        data.columns.tolist(),
    )

    if not as_frame:
        data = data.to_numpy()
        target = target.to_numpy()

    if return_X_y:
        return data, target

    diabetes_dataset = Bunch(
        data=data,
        target=target,
        feature_names=np.array(feature_names),
        frame=frame.assign(target=target),
        target_names=np.array(target_names),
    )

    return diabetes_dataset


# function to download thyroid dataset
def download_thyroid_dataset(path: Path = temp_dir) -> None:
    # download from url
    url = "https://archive.ics.uci.edu/static/public/102/thyroid+disease.zip"

    file_path = path.joinpath("thyroid.zip")

    urllib.request.urlretrieve(url, file_path)
    unzipped_path = path.joinpath("thyroid")

    # unzip
    with zipfile.ZipFile(file_path, "r") as zip_ref:
        zip_ref.extractall(unzipped_path)

    # remove zip file
    file_path.unlink()

    return None


# function to process the raw data and save it as a parquet file
def process_raw_and_save_thyroid(path: Path = temp_dir) -> None:

    file_path = path.joinpath("thyroid/ann-train.data")

    with open(file_path, "r") as f:
        lines = f.readlines()

    data = []
    for line in lines:
        data.append(line.strip().split(" "))

    data = pd.DataFrame(
        data,
        columns=[
            "age",
            "sex",
            "patient on thyroxine",
            "maybe on thyroxine",
            "on antithyroid medication",
            "patient reports malaise",
            "patient pregnant",
            "history of thyroid surgery",
            "patient on I131 treatment",
            "maybe hypothyroid",
            "maybe hyperthyroid",
            "patient on lithium",
            "patient has goitre",
            "patient has tumour",
            "patient hypopituitary",
            "psychological symptoms",
            "thyroid-stimulating hormone value (TSH)",
            "triiodothyronine value (T3)",
            "total thyroxine value (TT4)",
            "thyroxine uptake value (T4U)",
            "free thyroxine index value (FTI)",
            "target",
        ],
    )

    data = data.astype(
        {
            "age": "float",
            "sex": "int",
            "patient on thyroxine": "int",
            "maybe on thyroxine": "int",
            "on antithyroid medication": "int",
            "patient reports malaise": "int",
            "patient pregnant": "int",
            "history of thyroid surgery": "int",
            "patient on I131 treatment": "int",
            "maybe hypothyroid": "int",
            "maybe hyperthyroid": "int",
            "patient on lithium": "int",
            "patient has goitre": "int",
            "patient has tumour": "int",
            "patient hypopituitary": "int",
            "psychological symptoms": "int",
            "thyroid-stimulating hormone value (TSH)": "float",
            "triiodothyronine value (T3)": "float",
            "total thyroxine value (TT4)": "float",
            "thyroxine uptake value (T4U)": "float",
            "free thyroxine index value (FTI)": "float",
            "target": "int",
        }
    )

    data.to_parquet(path.joinpath("thyroid_dataset_original.parquet"))

    data["age"] = (data["age"] * 100).astype("int")

    class_of_interest = 1  # hypothyroid
    class_normal = 3

    data = data[data["target"].isin([class_of_interest, class_normal])]
    data["target"] = (data["target"] == class_of_interest).astype(int)

    # balanced sample from the data
    data = (
        data.groupby("target", group_keys=False)
        .sample(75, replace=False, random_state=0)
        .reset_index(drop=True)
    )

    columns_to_use_for_prediction = [
        "thyroid-stimulating hormone value (TSH)",
        "triiodothyronine value (T3)",
        "total thyroxine value (TT4)",
        "thyroxine uptake value (T4U)",
        # "free thyroxine index value (FTI)",
    ]

    data = data[columns_to_use_for_prediction + ["target"]]

    # save the processed data as a single file
    data.to_parquet(path.joinpath("thyroid_dataset.parquet"))

    return None


# function to clean up the files
def clean_thyroid_data_files(path: Path = temp_dir) -> None:

    # remove the unzipped folder
    shutil.rmtree(path.joinpath("thyroid"))

    return None


# function to download and process the thyroid dataset
def download_and_process_raw_thyroid(path: Path = temp_dir) -> None:

    print("Downloading thyroid dataset")
    download_thyroid_dataset(path)

    print("Processing and saving thyroid dataset")
    process_raw_and_save_thyroid(path)

    print("Cleaning up the files")
    clean_thyroid_data_files(path)

    return None


# function to load the thyroid dataset
def load_hypothyroid(
    as_frame: bool = False,
    return_X_y: bool = False,
    path: Path = temp_dir,
) -> Bunch:
    """
    Load and return the hypothyroid dataset.

    data: https://archive.ics.uci.edu/dataset/102/thyroid+disease
    feature and target information: https://arxiv.org/abs/cs/9503102

    Arguments
    ---------

    path: str or Path
        Path to the file where the data is stored or should be
        downloaded.
        Defaults to :code:`temp_dir`.

    return_X_y: bool
        Whether to return data and target as numpy arrays.
        Defaults to :code:`False`.

    as_frame: bool
        Whether to return the data as a pandas DataFrame.
        Defaults to :code:`False`.


    Returns
    -------

    Bunch
        Dictionary-like object, the interesting attributes are:
        'data', the data to learn, 'target', the classification target for each sample,
        'feature_names', the feature names, 'target_names', the target names,
        and 'frame', the pandas DataFrame.

    Tuple
        If :code:`return_X_y=True`, then a tuple of data and target.
        If :code:`as_frame=True`, then as pandas DataFrame otherwise as numpy arrays.

    """
    file_path = path.joinpath("thyroid_dataset.parquet")

    if not file_path.exists():
        download_and_process_raw_thyroid(path)

    thyroid_dataset = pd.read_parquet(file_path)

    data = thyroid_dataset.drop(columns="target")
    target = thyroid_dataset["target"]

    feature_names = data.columns.to_numpy()
    target_names = np.array(["normal", "hypothyroidism"])

    if not as_frame:
        data = data.to_numpy()
        target = target.to_numpy()

    if return_X_y:
        return data, target

    thyroid_dataset = Bunch(
        data=data,
        target=target,
        feature_names=np.array(feature_names),
        frame=thyroid_dataset,
        target_names=target_names,
    )

    return thyroid_dataset


def make_fake_data(
    y_fn: typing.Callable,
    n_samples: int,
    n_features: int,
    rng: np.random.Generator = np.random.default_rng(),
) -> pd.DataFrame:
    """
    This function generates fake data for testing purposes.

    Arguments
    ----------
    y_fn: callable
        A function that takes a 2D numpy array as input and returns a 1D numpy array.

    n_samples: int
        The number of samples to generate.

    n_features: int
        The number of features to generate.

    rng: np.random.Generator
        The random number generator to use.
        Defaults to :code:`np.random.default_rng()`.

    Returns
    -------
    X: pd.DataFrame
        The features.

    y: pd.Series
        The target.

    """
    X = pd.DataFrame(
        rng.normal(size=(n_samples, n_features)),
        columns=[f"feature {i}" for i in range(n_features)],
    )

    y = pd.Series(y_fn(X.values), name="target")

    return X, y


def save_fake_data(
    X: pd.DataFrame, y: pd.Series, path: Path = temp_dir
) -> None:
    """
    This function saves fake data to disk.
    This file will be saved in the directory
    specified by the path argument and
    called :code:`known_relationship.parquet`.

    Arguments
    ----------

    X: pd.DataFrame
        The features.

    y: pd.Series
        The target.

    path: str or Path
        The directory where the file should be saved.
        The file will be saved as :code:`known_relationship.parquet` in
        the specified directory.
        Defaults to :code:`temp_dir`.

    Returns
    -------

    None

    """

    path = path.joinpath("known_relationship.parquet")

    X.join(y).to_parquet(path, index=False)

    return None


def load_fake_data(
    path: Path = temp_dir,
    return_X_y: bool = False,
    as_frame: bool = False,
) -> Bunch:
    """
    Load and return the fake_data produced and saved using the
    functions :code:`make_fake_data` and :code:`save_fake_data`.

    Arguments
    ---------

    path: str or Path
        Path to the file where the data is stored or should be
        downloaded.
        Defaults to :code:`temp_dir`.

    return_X_y: bool
        Whether to return data and target as numpy arrays.
        Defaults to :code:`False`.

    as_frame: bool
        Whether to return the data as a pandas DataFrame.
        Defaults to :code:`False`.


    Returns
    -------

    Bunch
        Dictionary-like object, the interesting attributes are:
        'data', the data to learn, 'target', the classification target for each sample,
        'feature_names', the feature names, 'target_names', the target names,
        and 'frame', the pandas DataFrame.

    Tuple
        If :code:`return_X_y=True`, then a tuple of data and target.
        If :code:`as_frame=True`, then as pandas DataFrame otherwise as numpy arrays.


    """

    if Path(path).joinpath("known_relationship.parquet").exists():
        fake_data_dataset = pd.read_parquet(path.joinpath("known_relationship.parquet"))
    else:
        print("The fake data has not been generated yet. Generating it now.")
        rng = np.random.default_rng(42)
        
        # define the function that generates the target variable
        def y_fn(X):
            # y = 2 * x1 - 1 * x2 + 1 * x3 + noise
            true_y = 2 * X[:, 0] - 1 * X[:, 1] + X[:, 2]
            # some noise
            noise_y = 0.05 * rng.normal(size=(X.shape[0]))
            return true_y + noise_y


        # make fake data
        X, y = make_fake_data(y_fn, n_samples=250, n_features=3, rng=rng)

        # save the fake data
        save_fake_data(X, y)

        fake_data_dataset = pd.read_parquet(path.joinpath("known_relationship.parquet"))


    data = fake_data_dataset.drop(columns="target")
    target = fake_data_dataset["target"]

    feature_names = data.columns
    target_names = target.name

    if not as_frame:
        data = data.to_numpy()
        target = target.to_numpy()

    if return_X_y:
        return data, target

    fake_data_dataset = Bunch(
        data=data,
        target=target,
        feature_names=np.array(feature_names),
        frame=fake_data_dataset,
        target_names=np.array([target_names]),
    )

    return fake_data_dataset


def load_raw_dataset_frame(
    dataset_name: str = None,
    path: Path = temp_dir,
) -> pd.DataFrame:
    """
    Load the raw dataset, without any of the preprocessing done.
    This includes the original feature names and target names.

    This is helpful for assessing whether the language model has
    previously seen these datasets.

    Arguments
    ---------

    dataset_name: str
        The name of the dataset to load.
        This can be any of the following:
        - "fake_data"
        - "breast_cancer"
        - "diabetes"
        - "california_housing"
        - "diabetes_37"
        - "heart_disease"
        - "wine_quality"
        - "iris"
        - "wine"
        - "adult"
        - "hypothyroid"


    Returns
    -------

    pd.DataFrame
        The raw dataset.

    """

    if dataset_name == "fake_data":
        try:
            return pd.read_parquet(path.joinpath("known_relationship.parquet"))
        except FileNotFoundError:
            raise ValueError(
                """
                The fake data has not been generated yet. 
                Please run the make_fake_data.py script.
                """
            )

    elif dataset_name == "breast_cancer":
        return skd.load_breast_cancer(as_frame=True)["frame"]

    elif dataset_name == "diabetes":
        return skd.load_diabetes(as_frame=True, scaled=False)["frame"]

    elif dataset_name == "california_housing":
        return skd.fetch_california_housing(as_frame=True, data_home=path)["frame"]

    elif dataset_name == "diabetes_37":
        return skd.fetch_openml(data_id=37, as_frame=True, data_home=path)["frame"]

    # all of the following are loaded from the UCI repository
    elif dataset_name in [
        "heart_disease",
        "wine_quality",
        "iris",
        "wine",
        "adult",
    ]:
        return load_and_save_uci_original_dataset(dataset_name=dataset_name, path=path)

    elif dataset_name == "hypothyroid":
        return pd.read_parquet(path.joinpath("thyroid_dataset_original.parquet"))

    else:
        raise ValueError(f"Unknown dataset name: {dataset_name}")



from setuptools import setup

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='autoelicit',
    version='0.1.5',    
    description='A python package for eliciting prior knowledge from experts',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/alexcapstick/autoelicit',
    author='Alexander Capstick',
    author_email='alexander.capstick19@imperial.ac.uk',
    license='MIT',
    packages=['autoelicit'],
    install_requires=[
        "numpy",
        "scipy",
        "pandas",
        "scikit-learn",
        "tqdm",
        "pyarrow",
        "ucimlrepo",
        "openai>=1.51.2"        
    ],
    classifiers=[
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
)


"""
In this file, we implement two of the tests from:

Elephants Never Forget: Memorization and Learning 
of Tabular Data in Large Language Models 
(https://doi.org/10.48550/arXiv.2404.06209)

In particular, we implement the header completion test
and the row completion test. These tests are used to
determine if the language model has memorised the
header of the dataset and rows of the dataset
respectively.

We have adapted code from the original repository to
work with our implementation of the GPT-3 API and 
with pandas dataframes instead of csv files. We have
also provided further documentation for the functions and
removed one of the few-shot datasets from the header 
completion test as we use it as part of our experiments.

(https://github.com/interpretml/LLM-Tabular-Memorization-Checker)

The original repository is provided under the MIT license.

"""

import io
import typing as t
import numpy as np
import pandas as pd
import tqdm
from .datasets import load_raw_dataset_frame
from .gpt import LLMOutputs

# dataframe testing functions

header_system_prompt = """
You are an autocomplete bot for tabular datasets.
You will be prompted with parts of a tabular dataset.
Your task is to complete the dataset.
"""

row_system_prompt = """You are a helpful autocomplete 
bot for tabular datasets. Your task is to provide rows 
as they are contained in tabular datasets. The user 
provides a number of contiguous rows from a tabular 
dataset. You then provide the next row from the dataset.
"""


def df_to_string(df: pd.DataFrame) -> str:
    """
    Convert a pandas DataFrame to a string.
    """
    output = io.StringIO()
    df.to_csv(output, index=False)
    df_string = output.getvalue()
    output.close()
    return df_string


def header_completion_test(
    df: pd.DataFrame,
    client: LLMOutputs,
    split_rows: t.List[int] = [2, 4, 6, 8],
    completion_length: int = 500,
    system_prompt: str = header_system_prompt.replace("\n", " "),
    few_shot_datasets: t.List[str] = [
        "iris",
        "adult",
        "california_housing",
        "wine",
    ],
    rng: np.random.Generator = np.random.default_rng(),
) -> t.Tuple[str, str, str]:
    """
    Based on code given in:
    https://github.com/interpretml/LLM-Tabular-Memorization-Checker

    This function checks if the language model has
    memorised the first few rows of the dataset.

    The test is as follows:
    - The given dataframe will be converted into a string.
    - The string will be tructed at the :code:`split_row`
    row for each :code:`split_row` in :code:`split_rows`.
    - The language model will be asked to complete the next
    tokens, which will be measured against the next true
    :code:`completion_length` tokens.

    Few shot examples from the given datasets in :code:`few_shot_datasets`.

    Arguments
    ---------

    df: pd.DataFrame
        The dataset to check.

    client: LLMOutputs
        The client to use to generate completions.

    split_rows: List[int]
        The rows at which to split the dataset.
        Each integer is used for a separate experiment and
        determines the row at which the text is truncated
        and the language model asked to complete from.
        Defaults to :code:`[2, 4, 6, 8]`.

    completion_length: int
        The length of the completion to generate.
        Defaults to :code:`500`.

    system_prompt: str
        The prompt to use for the system.
        Defaults to the header_system_prompt.

    few_shot_files: List[str]
        The few-shot files to use. For the strings,
        the files will be loaded using the function
        :code:`load_raw_dataset_frame`. For the dataframes,
        they will be used directly.
        Defaults to :code:`["iris", "adult", "diabetes", "wine"]`.

    rng: np.random.Generator
        The random number generator to use.
        Defaults to :code:`np.random.default_rng()`.

    Returns
    -------

    Tuple[str, str, str]
        A tuple with the header prompt, the true completion, and the LLM completion.

    """

    few_shot_data = [
        (
            df_to_string(load_raw_dataset_frame(ds))
            if type(ds) == str  # only load the data if it is a string
            else df_to_string(ds)  # otherwise use the dataframe directly
        )
        for ds in few_shot_datasets
    ]

    df_string = df_to_string(df)
    df_rows_string = df_string.split("\n")

    # perform the test multiple times, cutting the dataset at
    # random positions in rows split_rows
    num_completions = -1
    header_prompt, llm_completion = None, None

    # for each of the row cuts
    for i_row in tqdm.tqdm(split_rows, desc="Testing header completion"):

        # find this row in the string
        offset = np.sum([len(row) for row in df_rows_string[: i_row - 1]])
        offset += rng.integers(
            len(df_rows_string[i_row]) // 3, 2 * len(df_rows_string[i_row]) // 3
        )
        # split the dataset at this row
        prefixes = [df_string[:offset]]
        # the completion is the next completion_length characters
        suffixes = [df_string[offset : offset + completion_length]]

        # construct the few-shot data: 1 example from each dataset
        few_shot = [
            ([fs_data[:offset]], [fs_data[offset : offset + completion_length]])
            for fs_data in few_shot_data
        ]

        test_prefixes = []
        test_suffixes = []
        responses = []

        # system prompt
        messages = [
            {
                "role": "system",
                "content": system_prompt,
            },
        ]

        # add the few-shot examples to the prompt
        for fs_prefixes, fs_suffixes in few_shot:
            fs_prefix, fs_suffix = None, None
            retries = 0
            # select a random prefix/suffix pair
            while (
                fs_prefix is None
                # assert that the test suffix is not contained in the
                # few-shot prefixes or suffixes
                or suffixes[0] in fs_prefix
                or suffixes[0] in fs_suffix
            ):
                fs_idx = rng.choice(len(fs_prefixes))
                fs_prefix = fs_prefixes[fs_idx]
                fs_suffix = fs_suffixes[fs_idx]
                retries += 1
                if retries > 100:
                    raise Exception(
                        """
                        Unable to construct a query where the desired output
                        is not contained in the few-shot data.
                        Did you provide the test dataset as few-shot example?
                        """
                    )
            messages.append({"role": "user", "content": fs_prefix})
            messages.append({"role": "assistant", "content": fs_suffix})

        # test observation
        test_prefix = prefixes[0]
        test_suffix = suffixes[0]
        messages.append({"role": "user", "content": test_prefix})

        # get the completion
        response = client.get_result(messages)

        # save the test data
        test_prefixes.append(test_prefix)
        test_suffixes.append(test_suffix)
        responses.append(response)

        # find the first digit where the response and the completion disagree
        idx = -1000
        for idx, (c, r) in enumerate(zip(df_string[offset:], response)):
            if c != r:
                break
        if idx == len(response) - 1 and response[idx] == df_string[offset + idx]:
            idx += 1  # no disagreement found, set idx to length of the response

        # is this the best completion so far?
        if idx > num_completions:
            num_completions = idx
            header_prompt = prefixes[0]
            llm_completion = response
            header_completion = df_string[offset : offset + len(llm_completion)]

    # return the best completion found
    test_triplet = header_prompt, header_completion, llm_completion

    return test_triplet


def row_completion_test(
    df: pd.DataFrame,
    client: LLMOutputs,
    n_prompt_rows: int = 10,
    n_tests: int = 25,
    n_fewshot_examples: int = 7,
    system_prompt: str = row_system_prompt.replace("\n", " "),
    rng: np.random.Generator = np.random.default_rng(),
) -> t.Tuple[str, str, str]:
    """
    Based on code given in:
    https://github.com/interpretml/LLM-Tabular-Memorization-Checker

    This function checks if the language model has
    memorised the header of the dataset.

    The test is as follows:
    - The given dataframe will be converted into a string.
    - :code:`n_tests` times:
        - Random chunks of :code:`n_prompt_rows` rows will be provided to the
        language model.
        - The language model will be asked to complete the next row.

    Few shot examples (in total :code:`n_fewshot_examples` examples)
    from the same dataset will be provided.

    Arguments
    ---------

    df: pd.DataFrame
        The dataset to check.

    client: LLMOutputs
        The client to use to generate completions.

    n_prompt_rows: int
        The number of rows to use as a prompt when
        testing whether the language model can complete the
        next row.
        Defaults to :code:`10`.

    n_tests: int
        The number of tests to run. For each
        test, a new row will be selected as the test row.
        Defaults to :code:`25`.

    n_fewshot_examples: int
        The number of few-shot examples to provide to the model.
        This will be chosen randomly from the dataset.
        Defaults to :code:`7`.

    system_prompt: str
        The prompt to use for the system.
        Defaults to the header_system_prompt.

    rng: np.random.Generator
        The random number generator to use.
        Defaults to :code:`np.random.default_rng()`.

    Returns
    -------

    test_triplet: Tuple[str, str, str]
        A tuple containing the test prefix, test suffix and the
        completion generated by the language model.

    """

    # getting df as string
    df_string = df_to_string(df)
    df_rows_string = df_string.split("\n")

    # calculate the proportion of duplicates
    duplicate_proportion = 1 - len(set(df_rows_string)) / len(df_rows_string)

    # prepare data
    prefixes = []
    suffixes = []

    # for each row in the df, make the prefixes from a sliding window
    # of n_prompt_rows rows and the suffixes the next row
    for idx in range(len(df_rows_string) - n_prompt_rows):
        prefixes.append("\n".join(df_rows_string[idx : idx + n_prompt_rows]))
        suffixes.append(df_rows_string[idx + n_prompt_rows])

    # shuffle the data
    idx = rng.permutation(len(prefixes))
    prefixes = [prefixes[i] for i in idx]
    suffixes = [suffixes[i] for i in idx]

    # the number of points to evaluate
    num_points = min(n_tests, len(prefixes))

    test_prefixes = []
    test_suffixes = []
    llm_completion = []

    # testing each point one at a time
    for i_testpoint in tqdm.trange(num_points, desc="Testing row completion"):
        # system prompt
        messages = [
            {
                "role": "system",
                "content": system_prompt,
            },
        ]

        # add the few-shot examples
        for _ in range(n_fewshot_examples):
            idx = None
            retries = 0
            # select a random prefix/suffix pair
            while (
                idx is None
                # assert that the demonstration
                # is not the test point
                or idx == i_testpoint
                # assert that the current test suffix is not
                # contained in the few-shot prefixes or suffixes
                or suffixes[i_testpoint] in prefixes[idx]
                or suffixes[i_testpoint] in suffixes[idx]
            ):
                idx = rng.choice(len(prefixes))
                retries += 1
                if retries > 100:
                    raise Exception(
                        """Unable to construct a query where the desired 
                        output is not contained in the few-shot data.
                        Did you provide the test dataset as few-shot example?"""
                    )

            # if the above is met then add the few-shot example
            # to the prompt
            prefix = prefixes[idx]
            suffix = suffixes[idx]
            messages.append({"role": "user", "content": prefix})
            messages.append({"role": "assistant", "content": suffix})

        # test observation
        test_prefix = prefixes[i_testpoint]
        test_suffix = suffixes[i_testpoint]
        messages.append({"role": "user", "content": test_prefix})

        # get the completion
        response = client.get_result(messages)

        # store prefix, suffix and response
        test_prefixes.append(test_prefix)
        test_suffixes.append(test_suffix)
        llm_completion.append(response)

    test_triplet = list(zip(test_prefixes, test_suffixes, llm_completion))

    return test_triplet



import copy
import typing as t
import numpy as np
import pandas as pd
import difflib
from pathlib import Path


def inv_logistic(x: np.array) -> np.array:
    """
    The inverse of the logistic function.

    Arguments
    ----------

    x : array-like
        The input to the inverse logistic function.

    Returns
    ---------

    y : array-like
        The output of the inverse logistic function.

    """
    x = np.array(x)
    return np.log(x / (1 - x))


def load_prompts(path: Path, delim="\n\n") -> t.List[str]:
    """
    This function loads a text file of prompts into a list.
    In the text file, each prompt should be separated by
    a delimiter that is specified by the :code:`delim` argument.

    Arguments
    ----------
    path : Path
        The path to the text file containing the prompts.

    delim : str
        The delimiter that separates the prompts in the text file.
        Defaults to :code:`\n\n`.

    Returns
    --------

    prompts : list of str
        A list of prompts loaded from the text file.


    """
    return open(path).read().split(delim)


def load_nested_dict_to_pandas(
    results_dict: t.Dict[t.Any, t.Any],
    level_names: t.List[t.Any] = None,
) -> pd.DataFrame:
    """
    This function loads a nested dictionary into a pandas DataFrame.

    Arguments
    ----------

    results_dict : dict
        The nested dictionary to load.

    level_names : list
        The names of the levels in the dictionary.
        This can be used to name the columns of the
        outputted :code:`DataFrame`.
        Defaults to :code:`None`.


    Returns
    --------

    df : DataFrame
        The pandas DataFrame with the nested dictionary loaded.


    """

    # internal function to call recursively
    def internal_load_nested_dict_to_pandas(results_dict, level, level_names):
        df = pd.DataFrame()
        level = level + 1
        for key in results_dict.keys():
            if isinstance(results_dict[key], dict):
                df_output = internal_load_nested_dict_to_pandas(
                    results_dict[key], level, level_names
                )
                if level_names is None:
                    df_output.insert(0, f"level_{level}", key)
                else:
                    df_output.insert(0, level_names[level], key)
                df = pd.concat([df, df_output])
            else:
                try:
                    df[key] = [results_dict[key]]
                except:
                    df[key] = np.nan
        return df

    return internal_load_nested_dict_to_pandas(
        results_dict=results_dict, level=-1, level_names=level_names
    )


def cut_end_points(
    x: np.array,
    lower_percentile: float,
    upper_percentile: float,
) -> np.array:
    """
    Return an array that is a subset of :code:`x` that
    only contains data from with in the :code:`lower_percentile`
    and :code:`upper_percentile`.


    Arguments
    ----------

    - x : array-like
        The input array.

    - lower_percentile : float
        The lower percentile to cut the data at.
        This should be in the range [0, 100].

    - upper_percentile : float
        The upper percentile to cut the data at.
        This should be in the range [0, 100].

    Returns
    ---------
    - x_subset : array-like
        The subset of the input array that is within the
        :code:`lower_percentile` and :code:`upper_percentile`.

    """
    lower = np.percentile(x, lower_percentile)
    upper = np.percentile(x, upper_percentile)
    return x[(lower < x) & (x < upper)]


def cut_end_points_groupby(
    df: pd.DataFrame,
    groupby: str,
    cut_on: str,
    lower_percentile: float,
    upper_percentile: float,
) -> pd.DataFrame:
    """
    Return a dataframe that is a subset of :code:`df` that
    only contains data from with in the :code:`lower_percentile`
    and :code:`upper_percentile` in the :code:`cut_on` column
    grouped by the :code:`groupby` column.


    Arguments
    ----------

    - df : pd.DataFrame
        The input dataframe.

    - groupby : str
        The column to group the data by.

    - cut_on : str
        The column to cut the data on.

    - lower_percentile : float
        The lower percentile to cut the data at.
        This should be in the range [0, 100].

    - upper_percentile : float
        The upper percentile to cut the data at.
        This should be in the range [0, 100].

    Returns
    ---------
    - df_subset : array-like
        The subset of the input df that is within the
        :code:`lower_percentile` and :code:`upper_percentile`.

    """
    return df.loc[
        lambda x: x.groupby(groupby)[cut_on]
        .apply(
            lambda l: l.loc[
                (l > l.quantile(lower_percentile / 100))
                & (l < l.quantile(upper_percentile / 100))
            ]
        )
        .index.get_level_values(-1)
    ]


def make_list(x: t.Iterable) -> t.List[t.Any]:
    """
    This function will take an input and
    if any of its nested objects are arrays, they will be
    converted to lists.

    Arguments
    ----------

    x : any
        The input to convert to a list.

    Returns
    ---------

    x : list
        The input as a list.

    """

    # if x is an array then we can convert it straight away
    if isinstance(x, np.ndarray):
        x_new = x.tolist()
        return x_new

    # otherwise we need to find all the arrays in the object
    x_new = []
    for xi in x:
        if isinstance(xi, np.ndarray):
            xi = xi.tolist()
        if isinstance(xi, list):
            xi = make_list(xi)
        x_new.append(xi)
    return x_new


def find_best_matches(list1, list2):
    list1 = copy.deepcopy(list1)
    list2 = copy.deepcopy(list2)
    matches = []
    for item1 in list1:
        best_match = None
        highest_similarity = 0
        for item2 in list2:
            similarity = difflib.SequenceMatcher(None, item1, item2).ratio()
            if similarity > highest_similarity:
                highest_similarity = similarity
                best_match = item2
        matches.append((item1, best_match))
        list2.remove(best_match)  # Ensure each item is matched only once
    return matches
